"""Page Hinkley module."""

from typing import Optional, Union

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.change_detection.base import (
    BaseCUSUM,
    BaseCUSUMConfig,
    DeltaConfig,
    AlphaConfig,
)


class PageHinkleyConfig(BaseCUSUMConfig, DeltaConfig, AlphaConfig):
    """Page Hinkley [page1954continuous]_ configuration.

    :param delta: delta value, defaults to 0.005
    :type delta: float
    :param lambda_: lambda value, defaults to 50.0
    :type lambda_: float
    :param alpha: forgetting factor value, defaults to 0.9999
    :type alpha: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [page1954continuous] Page, Ewan S.
        "Continuous inspection schemes."
        Biometrika 41.1/2 (1954): 100-115.
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        delta: float = 0.005,
        lambda_: float = 50.0,
        alpha: float = 0.9999,
        min_num_instances: int = 30,
    ) -> None:
        BaseCUSUMConfig.__init__(
            self, min_num_instances=min_num_instances, lambda_=lambda_
        )
        DeltaConfig.__init__(self, delta=delta)
        AlphaConfig.__init__(self, alpha=alpha)


class PageHinkley(BaseCUSUM):
    """Page Hinkley [page1954continuous]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`PageHinkleyConfig` is used.
    :type config: Optional[PageHinkleyConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :References:

    .. [page1954continuous] Page, Ewan S.
        "Continuous inspection schemes."
        Biometrika 41.1/2 (1954): 100-115.

    :Example:

    >>> from frouros.detectors.concept_drift import PageHinkley
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.normal(loc=0.2, scale=0.01, size=1000)
    >>> dist_b = np.random.normal(loc=0.8, scale=0.04, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = PageHinkley()
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    Change detected at step 1094
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = PageHinkleyConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[PageHinkleyConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )

    def _update_sum(self, error_rate: float) -> None:
        self.sum_ = self.config.alpha * self.sum_ + (  # type: ignore
            error_rate - self.mean_error_rate.mean - self.config.delta  # type: ignore
        )



"""Concept drift SPC (statistical process control) detection methods' init."""

from .ddm import DDM, DDMConfig
from .ecdd import ECDDWT, ECDDWTConfig
from .eddm import EDDM, EDDMConfig
from .hddm import HDDMA, HDDMAConfig, HDDMW, HDDMWConfig
from .rddm import RDDM, RDDMConfig

__all__ = [
    "DDM",
    "DDMConfig",
    "ECDDWT",
    "ECDDWTConfig",
    "EDDM",
    "EDDMConfig",
    "HDDMA",
    "HDDMAConfig",
    "HDDMW",
    "HDDMWConfig",
    "RDDM",
    "RDDMConfig",
]



"""Real datasets module."""

from typing import Optional

import numpy as np  # type: ignore
from scipy.io import arff  # type: ignore

from frouros.datasets.base import BaseDatasetDownload


class Elec2(BaseDatasetDownload):
    """Elec2 dataset [harries1999splice]_.

    :param file_path: file path for the downloaded file, defaults to None. If None, the file will be downloaded to a temporary file.
    :type file_path: Optional[str]

    :Note:
    Dataset can be downloaded from the following sources (in order of preference):

    - https://nextcloud.ifca.es/index.php/s/2coqgBEpa82boLS/download
    - https://www.openml.org/data/download/2419/electricity-normalized.arff

    :References:

    .. [harries1999splice] Harries, Michael.
        "Splice-2 comparative evaluation: Electricity pricing." (1999).

    :Example:

    >>> from frouros.datasets.real import Elec2
    >>> elec2 = Elec2()
    >>> elec2.download()
    INFO:frouros:Trying to download data from https://nextcloud.ifca.es/index.php/s/2coqgBEpa82boLS/download to /tmp/tmpro3ienx0
    >>> dataset = elec2.load()
    >>> dataset
    array([(0.    , b'2', 0.      , 0.056443, 0.439155, 0.003467, 0.422915, 0.414912, b'UP'),
       (0.    , b'2', 0.021277, 0.051699, 0.415055, 0.003467, 0.422915, 0.414912, b'UP'),
       (0.    , b'2', 0.042553, 0.051489, 0.385004, 0.003467, 0.422915, 0.414912, b'UP'),
       ...,
       (0.9158, b'7', 0.957447, 0.043593, 0.34097 , 0.002983, 0.247799, 0.362281, b'DOWN'),
       (0.9158, b'7', 0.978723, 0.066651, 0.329366, 0.00463 , 0.345417, 0.206579, b'UP'),
       (0.9158, b'7', 1.      , 0.050679, 0.288753, 0.003542, 0.355256, 0.23114 , b'DOWN')],
      dtype=[('date', '<f8'), ('day', 'S1'), ('period', '<f8'), ('nswprice', '<f8'), ('nswdemand', '<f8'), ('vicprice', '<f8'), ('vicdemand', '<f8'), ('transfer', '<f8'), ('class', 'S4')])
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        file_path: Optional[str] = None,
    ) -> None:
        super().__init__(
            url=[
                "https://nextcloud.ifca.es/index.php/s/2coqgBEpa82boLS/download",
                "https://www.openml.org/data/download/2419/electricity-normalized.arff",
            ],
            file_path=file_path,
        )

    def read_file(self, **kwargs) -> np.ndarray:
        """Read file.

        :param kwargs: dict of kwargs
        :type kwargs: dict
        :return: read file
        :rtype: numpy.ndarray
        """
        index = kwargs.get("index", 0)
        dataset = arff.loadarff(f=self.file_path)[index]
        return dataset



"""JS (Jensen-Shannon distance) module."""

from typing import Any, Optional, Union

import numpy as np  # type: ignore
from scipy.spatial.distance import jensenshannon  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBasedProbability,
    DistanceResult,
)


class JS(BaseDistanceBasedProbability):
    """JS (Jensen-Shannon distance) [lin1991divergence]_ detector.

    :param num_bins: number of bins in which to divide probabilities, defaults to 10
    :type num_bins: int
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]
    :param kwargs: additional keyword arguments to pass to scipy.spatial.distance.jensenshannon
    :type kwargs: dict[str, Any]

    :References:

    .. [lin1991divergence] Lin, Jianhua.
        "Divergence measures based on the Shannon entropy."
        IEEE Transactions on Information theory 37.1 (1991): 145-151.

    :Example:

    >>> from frouros.detectors.data_drift import JS
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = JS(num_bins=20)
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    DistanceResult(distance=0.41702877367162156)
    """  # noqa: E501

    def __init__(  # noqa: D107
        self,
        num_bins: int = 10,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            statistical_method=self._js,
            statistical_kwargs={
                "num_bins": num_bins,
                **kwargs,
            },
            callbacks=callbacks,
        )
        self.num_bins = num_bins
        self.kwargs = kwargs

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        js = self._js(X=X_ref, Y=X, num_bins=self.num_bins, **self.kwargs)
        distance = DistanceResult(distance=js)
        return distance

    @staticmethod
    def _js(
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
        *,
        num_bins: int,
        **kwargs: dict[str, Any],
    ) -> float:
        (  # noqa: N806
            X_ref_rvs,
            X_rvs,
        ) = BaseDistanceBasedProbability._calculate_probabilities(
            X_ref=X,
            X=Y,
            num_bins=num_bins,
        )
        js = jensenshannon(p=X_ref_rvs, q=X_rvs, **kwargs)
        return js



"""Base detector module."""

import abc
from typing import Any, Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.base import BaseCallback


class BaseDetector(abc.ABC):
    """Abstract class representing a detector."""

    def __init__(
        self,
        callbacks: Optional[Union[BaseCallback, list[BaseCallback]]] = None,
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallback, list[Callback]]]
        """
        self.callbacks = callbacks  # type: ignore

    @property
    def callbacks(self) -> Optional[list[BaseCallback]]:
        """Callbacks property.

        :return: callbacks
        :rtype: Optional[list[BaseCallback]]
        """
        return self._callbacks  # type: ignore

    @callbacks.setter
    def callbacks(
        self,
        value: Optional[Union[BaseCallback, list[BaseCallback]]],
    ) -> None:
        """Callbacks setter.

        :param value: value to be set
        :type value: Optional[Union[BaseCallback, list[Callback]]]
        :raises TypeError: Type error exception
        """
        if value is not None:
            if isinstance(value, BaseCallback):
                self._callbacks = [value]
            elif not all(
                isinstance(callback, BaseCallback) for callback in value  # type: ignore
            ):
                raise TypeError("value must be of type None or a list of BaseCallback.")
            else:
                self._callbacks = value  # type: ignore
        else:
            self._callbacks = []

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset method."""

    def _get_callbacks_logs(self) -> dict[str, Any]:
        logs = {
            callback.name: callback.logs for callback in self.callbacks  # type: ignore
        }
        return logs

    @staticmethod
    def _check_array(X: Any) -> None:  # noqa: N803
        if not isinstance(X, np.ndarray):
            raise TypeError("X must be a numpy array")

    def __repr__(self) -> str:
        """Repr method.

        :return: repr value
        :rtype: str
        """
        return (
            f"{self.__class__.__name__}"
            f"(callbacks=["
            f"{', '.join([*map(str, self.callbacks)])}])"  # type: ignore
        )



"""Base data drift batch module."""

import abc
from typing import Any, Optional, Tuple, Union

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import (
    BaseDataDrift,
    BaseDataType,
    BaseStatisticalType,
)
from frouros.detectors.data_drift.exceptions import (
    MismatchDimensionError,
)
from frouros.utils.checks import check_callbacks


class BaseDataDriftBatch(BaseDataDrift):
    """Abstract class representing a data drift batch detector."""

    def __init__(
        self,
        data_type: BaseDataType,
        statistical_type: BaseStatisticalType,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        """Init method.

        :param data_type: data type
        :type data_type: BaseDataType
        :param statistical_type: statistical type
        :type statistical_type: BaseStatisticalType
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallbackBatch], list[BaseCallbackBatch]]
        """
        check_callbacks(
            callbacks=callbacks,
            expected_cls=BaseCallbackBatch,  # type: ignore
        )
        super().__init__(
            callbacks=callbacks,  # type: ignore
            data_type=data_type,
            statistical_type=statistical_type,
        )
        for callback in self.callbacks:  # type: ignore
            callback.set_detector(detector=self)

    def _fit(
        self,
        X: np.ndarray,  # noqa: N803
    ) -> None:
        self.X_ref = X  # type: ignore

    def compare(
        self,
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> Tuple[np.ndarray, dict[str, Any]]:
        """Compare values.

        :param X: test data
        :type X: numpy.ndarray
        :return: compare result and callbacks logs
        :rtype: Tuple[numpy.ndarray, dict[str, Any]]
        """
        for callback in self.callbacks:  # type: ignore
            callback.on_compare_start(  # type: ignore
                X_ref=self.X_ref,
                X_test=X,
            )
        result = self._compare(X=X, **kwargs)
        for callback in self.callbacks:  # type: ignore
            callback.on_compare_end(  # type: ignore
                result=result,
                X_ref=self.X_ref,
                X_test=X,
            )

        callbacks_logs = self._get_callbacks_logs()
        return result, callbacks_logs

    def _check_compare_dimensions(self, X: np.ndarray) -> None:  # noqa: N803
        try:
            if self.X_ref.shape[1] != X.shape[1]:  # type: ignore
                raise MismatchDimensionError(
                    f"Dimensions of X_ref ({self.X_ref.shape[-1]}) "  # type: ignore
                    f"and X ({X.shape[-1]}) must be equal"
                )
        except IndexError as e:
            if self.X_ref.ndim != X.ndim:  # type: ignore
                raise MismatchDimensionError(f"Dimensions of X ({X.ndim})") from e

    def _specific_checks(self, X: np.ndarray) -> None:  # noqa: N803
        self._check_compare_dimensions(X=X)

    @abc.abstractmethod
    def _apply_method(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> Any:
        pass

    @abc.abstractmethod
    def _compare(
        self,
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> np.ndarray:
        pass

    def _get_result(
        self,
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> Union[list[float], list[Tuple[float, float]], Tuple[float, float]]:
        result = self._apply_method(  # type: ignore # pylint: disable=not-callable
            X_ref=self.X_ref,
            X=X,
            **kwargs,
        )
        return result



"""Welch's t-test module."""

from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.stats import ttest_ind  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)


class WelchTTest(BaseStatisticalTest):
    """Welch's t-test [welch1947generalization]_ detector.

    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :Note:
    - Passing additional arguments to `scipy.stats.ttest_ind <https://docs.scipy.org/doc/scipy/reference/generated/scipy.stats.ttest_ind.html>`__ can be done using :func:`compare` kwargs.

    :References:

    .. [welch1947generalization] Welch, Bernard L.
        "The generalization of ‘STUDENT'S’problem when several different population
        varlances are involved."
        Biometrika 34.1-2 (1947): 28-35.

    :Example:

    >>> from frouros.detectors.data_drift import WelchTTest
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = WelchTTest()
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    StatisticalResult(statistic=-7.651304662806378, p_value=8.685225410826823e-13)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    @staticmethod
    def _statistical_test(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> StatisticalResult:
        test = ttest_ind(
            a=X_ref,
            b=X,
            equal_var=False,
            alternative=kwargs.get("alternative", "two-sided"),
            **kwargs,
        )
        test = StatisticalResult(
            statistic=test.statistic,
            p_value=test.pvalue,
        )
        return test



"""Kuiper's test module."""

from typing import Optional, Union

import numpy as np  # type: ignore
from scipy.special import comb, factorial  # type: ignore
from scipy.stats import ks_2samp  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)


class KuiperTest(BaseStatisticalTest):
    """Kuiper's test [kuiper1960tests]_ detector.

    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :References:

    .. [kuiper1960tests] Kuiper, Nicolaas H.
        "Tests concerning random points on a circle."
        Nederl. Akad. Wetensch. Proc. Ser. A. Vol. 63. No. 1. 1960.

    :Example:

    >>> from frouros.detectors.data_drift import KuiperTest
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = KuiperTest()
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    StatisticalResult(statistic=0.55, p_value=5.065664859580971e-13)
    """  # noqa: E501  # pylint: disable=line-too-long

    def __init__(  # noqa: D107
        self,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    @staticmethod
    def _statistical_test(
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,
        **kwargs,
    ) -> StatisticalResult:
        statistic, p_value = KuiperTest._kuiper(
            X=X_ref,
            Y=X,
            **kwargs,
        )
        test = StatisticalResult(
            statistic=statistic,
            p_value=p_value,
        )
        return test

    @staticmethod
    def _false_positive_probability(
        D: float,  # noqa: N803
        N: float,
    ) -> float:
        """
        Compute the false positive probability for the Kuiper's test.

        NOTE: This function is a modified version of the
        `kuiper_false_positive_probability` function from the
        `astropy.stats` <https://docs.astropy.org/en/stable/api/astropy.stats.kuiper_false_positive_probability.html#astropy.stats.kuiper_false_positive_probability> package.  # noqa: E501

        :param D: Kuiper's statistic
        :param D: float
        :param N: effective size of the sample
        :type N: float
        :return: false positive probability
        :rtype: float
        """
        if D < 2.0 / N:
            return 1.0 - factorial(N) * (D - 1.0 / N) ** (N - 1)

        if D < 3.0 / N:
            k = -(N * D - 1.0) / 2.0
            r = np.sqrt(k**2 - (N * D - 2.0) ** 2 / 2.0)
            a, b = -k + r, -k - r
            return 1 - (
                factorial(N - 1)
                * (b ** (N - 1) * (1 - a) - a ** (N - 1) * (1 - b))
                / N ** (N - 2)
                / (b - a)
            )

        if (D > 0.5 and N % 2 == 0) or (D > (N - 1.0) / (2.0 * N) and N % 2 == 1):
            # NOTE: the upper limit of this sum is taken from Stephens 1965
            t = np.arange(np.floor(N * (1 - D)) + 1)
            y = D + t / N
            Tt = y ** (t - 3) * (  # noqa: N806
                y**3 * N
                - y**2 * t * (3 - 2 / N)
                + y * t * (t - 1) * (3 - 2 / N) / N
                - t * (t - 1) * (t - 2) / N**2
            )
            term1 = comb(N, t)
            term2 = (1 - D - t / N) ** (N - t - 1)
            # term1 is formally finite, but is approximated by numpy as np.inf for
            # large values, so we set them to zero manually when they would be
            # multiplied by zero anyway
            term1[(term1 == np.inf) & (term2 == 0)] = 0.0
            return (Tt * term1 * term2).sum()

        z = D * np.sqrt(N)
        # When m*z>18.82 (sqrt(-log(finfo(double))/2)), exp(-2m**2z**2)
        # underflows.  Cutting off just before avoids triggering a (pointless)
        # underflow warning if `under="warn"`.
        ms = np.arange(1, 18.82 / z)
        S1 = (  # noqa: N806
            2 * (4 * ms**2 * z**2 - 1) * np.exp(-2 * ms**2 * z**2)
        ).sum()
        S2 = (  # noqa: N806
            ms**2 * (4 * ms**2 * z**2 - 3) * np.exp(-2 * ms**2 * z**2)
        ).sum()
        return S1 - 8 * D / 3 * S2

    @staticmethod
    def _kuiper(
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
    ) -> tuple[float, float]:
        """Kuiper's test.

        :param X: reference data
        :type X: numpy.ndarray
        :param Y: test data
        :type Y: numpy.ndarray
        :return: Kuiper's statistic and p-value
        :rtype: tuple[float, float]
        """
        X = np.sort(X)  # noqa: N806
        Y = np.sort(Y)  # noqa: N806
        X_size = len(X)  # noqa: N806
        Y_size = len(Y)  # noqa: N806

        statistic = ks_2samp(
            data1=X,
            data2=Y,
            alternative="two-sided",
        ).statistic
        sample_effective_size = X_size * Y_size / float(X_size + Y_size)

        p_value = KuiperTest._false_positive_probability(
            D=statistic,
            N=sample_effective_size,
        )

        return statistic, p_value



"""PSI (Population Stability Index) module."""

import sys
from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBasedBins,
    DistanceResult,
)


class PSI(BaseDistanceBasedBins):
    """PSI (Population Stability Index) [wu2010enterprise]_ detector.

    :param num_bins: number of bins in which to divide probabilities, defaults to 10
    :type num_bins: int
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :References:

    .. [wu2010enterprise] Wu, Desheng, and David L. Olson.
        "Enterprise risk management: coping with model risk in a large bank."
        Journal of the Operational Research Society 61.2 (2010): 179-190.

    :Example:

    >>> from frouros.detectors.data_drift import PSI
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = PSI(num_bins=20)
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    DistanceResult(distance=134.95409065116183)
    """

    def __init__(  # noqa: D107
        self,
        num_bins: int = 10,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            statistical_method=self._psi,
            statistical_kwargs={
                "num_bins": num_bins,
            },
            callbacks=callbacks,
        )
        self.num_bins = num_bins

    def _apply_method(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> DistanceResult:
        distance = self._distance_measure(X_ref=X_ref, X=X, **kwargs)
        return distance

    def _distance_measure_bins(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
    ) -> float:
        psi = self._psi(X=X_ref, Y=X, num_bins=self.num_bins)
        return psi

    @staticmethod
    def _psi(
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,  # noqa: N803
        num_bins: int,
    ) -> float:
        (  # noqa: N806
            X_percents,
            Y_percents,
        ) = BaseDistanceBasedBins._calculate_bins_values(
            X_ref=X, X=Y, num_bins=num_bins
        )
        # Replace 0.0 values with the smallest number possible
        # in order to avoid division by zero
        X_percents[X_percents == 0.0] = sys.float_info.min
        Y_percents[Y_percents == 0.0] = sys.float_info.min
        psi = np.sum((Y_percents - X_percents) * np.log(Y_percents / X_percents))
        return psi



"""HI (Histogram intersection) normalized complement module."""

from typing import Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.batch.base import BaseCallbackBatch
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBasedBins,
)


class HINormalizedComplement(BaseDistanceBasedBins):
    """HI (Histogram intersection) normalized complement [swain1991color]_ detector.

    :param num_bins: number of bins in which to divide probabilities, defaults to 10
    :type num_bins: int
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]]

    :References:

    .. [swain1991color] Swain, M. J., and D. H. Ballard.
        "Color Indexing International Journal of Computer
        Vision 7." (1991): 11-32.

    :Example:

    >>> from frouros.detectors.data_drift import HINormalizedComplement
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> X = np.random.normal(loc=0, scale=1, size=100)
    >>> Y = np.random.normal(loc=1, scale=1, size=100)
    >>> detector = HINormalizedComplement(num_bins=20)
    >>> _ = detector.fit(X=X)
    >>> detector.compare(X=Y)[0]
    DistanceResult(distance=0.53)
    """

    def __init__(  # noqa: D107
        self,
        num_bins: int = 10,
        callbacks: Optional[Union[BaseCallbackBatch, list[BaseCallbackBatch]]] = None,
    ) -> None:
        super().__init__(
            statistical_method=self._hi_normalized_complement,
            statistical_kwargs={
                "num_bins": num_bins,
            },
            callbacks=callbacks,
        )
        self.num_bins = num_bins

    def _distance_measure_bins(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
    ) -> float:
        intersection_normalized_complement = self._hi_normalized_complement(
            X=X_ref, Y=X, num_bins=self.num_bins
        )
        return intersection_normalized_complement

    @staticmethod
    def _hi_normalized_complement(
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
        *,
        num_bins: int,
    ) -> float:
        hist_range = (
            np.min([np.min(X), np.min(Y)]),
            np.max([np.max(X), np.max(Y)]),
        )
        X_hist, _ = np.histogram(  # noqa: N806
            X, bins=num_bins, range=hist_range  # noqa: N806
        )
        X_hist = X_hist / X.shape[0]  # noqa: N806
        Y_hist, _ = np.histogram(Y, bins=num_bins, range=hist_range)  # noqa: N806
        Y_hist = Y_hist / Y.shape[0]  # noqa: N806
        intersection_normalized_complement = 1 - np.sum(np.minimum(X_hist, Y_hist))

        return intersection_normalized_complement



"""BOCD (Bayesian Online Change Detection) module."""

import abc
import copy
from typing import Union, Optional

import numpy as np  # type: ignore
from scipy.special import logsumexp  # type: ignore
from scipy.stats import norm  # type: ignore

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.detectors.concept_drift.streaming.change_detection.base import (
    BaseChangeDetection,
    BaseChangeDetectionConfig,
)


class BaseBOCDModel(abc.ABC):
    """Abstract class representing a BOCD model."""

    @abc.abstractmethod
    def log_pred_prob(self, idx: int, value: Union[int, float]) -> float:
        """Log predictive probability.

        :param idx: index
        :type idx: int
        :param value: value
        :type value: Union[int, float]
        :return: log predictive probability
        :rtype: float
        """

    @abc.abstractmethod
    def update(self, value: Union[int, float], **kwargs) -> None:
        """Update model.

        :param value: value
        :type value: Union[int, float]
        """


class GaussianUnknownMean(BaseBOCDModel):
    """Gaussian unknown mean model.

    :Note:
     Adapted from the implementation in https://github.com/gwgundersen/bocd.
    """

    def __init__(
        self,
        prior_mean: float = 0,
        prior_var: float = 1,
        data_var: float = 1,
    ) -> None:
        """Init method.

        :param prior_mean: prior mean
        :type prior_mean: float
        :param prior_var: prior variance
        :type prior_var: float
        :param data_var: data variance
        :type data_var: float
        """
        super().__init__()
        self.mean_params = np.array([prior_mean])
        self.precision_params = np.array([1 / prior_var])
        self.data_var = data_var

    @property
    def data_var(self) -> float:
        """Data variance property.

        :return: data variance
        :rtype: float
        """
        return self._data_var

    @data_var.setter
    def data_var(self, value: float) -> None:
        """Data variance setter.

        :param value: value to be set
        :type value: float
        :raises ValueError: Value error exception
        """
        if value <= 0:
            raise ValueError("data_var must be greater than 0.")
        self._data_var = value

    def log_pred_prob(self, idx: int, value: Union[int, float]) -> float:
        """Log predictive probability.

        :param idx: index
        :type idx: int
        :param value: value
        :type value: Union[int, float]
        :return: log predictive probability
        :rtype: float
        """
        post_means = self.mean_params[:idx]
        post_stds = np.sqrt(self.var_params[:idx])
        return norm(post_means, post_stds).logpdf(value)

    def update(self, value: Union[int, float], **kwargs) -> None:
        """Update model.

        :param value: value
        :type value: Union[int, float]
        """
        new_param_prec = self.precision_params + (1 / self.data_var)
        self.precision_params = np.append([self.precision_params[0]], new_param_prec)
        new_param_mean = (
            self.mean_params * self.precision_params[:-1] + (value / self.data_var)
        ) / new_param_prec
        self.mean_params = np.append([self.mean_params[0]], new_param_mean)

    @property
    def var_params(self):
        """Helper function for computing the posterior variance."""
        return 1 / self.precision_params + self.data_var


class BOCDConfig(BaseChangeDetectionConfig):
    """BOCD (Bayesian Online Change Detection) [adams2007bayesian]_ configuration.

    :param model: BOCD model, defaults to None. If None, :class:`frouros.detectors.concept_drift.streaming.change_detection.bocd.GaussianUnknownMean` is used.
    :type model: Optional[BaseBOCDModel]
    :param hazard: hazard value, defaults to 0.01
    :type hazard: float
    :param min_num_instances: minimum numbers of instances to start looking for changes, defaults to 30
    :type min_num_instances: int

    :References:

    .. [adams2007bayesian] Adams, Ryan Prescott, and David JC MacKay.
        "Bayesian online changepoint detection."
        arXiv preprint arXiv:0710.3742 (2007).
    """  # noqa: E501  pylint: disable=line-too-long

    model_type = GaussianUnknownMean

    def __init__(  # noqa: D107
        self,
        model: Optional[BaseBOCDModel] = None,  # type: ignore
        hazard: float = 0.01,
        min_num_instances: int = 30,
    ) -> None:
        super().__init__(
            min_num_instances=min_num_instances,
        )
        self.model = model  # type: ignore
        self.log_hazard = np.log(hazard)
        self.log_1_minus_hazard = np.log(1 - hazard)

    @property
    def model(self) -> BaseBOCDModel:
        """Get model.

        :return: model
        :rtype: BaseBOCDModel
        """
        return self._model

    @model.setter
    def model(self, model: BaseBOCDModel) -> None:
        """Set model.

        :param model: model
        :type model: BaseBOCDModel
        :raises TypeError: if model is not an instance of BaseModel
        """
        if model is not None:
            if not isinstance(model, BaseBOCDModel):
                raise TypeError(
                    f"model must be an instance of BaseModel, not {type(model)}"
                )
            self._model = model
        else:
            self._model = self.model_type()


class BOCD(BaseChangeDetection):
    """BOCD (Bayesian Online Change Detection) [adams2007bayesian]_ detector.

    :param config: configuration object of the detector, defaults to None. If None, the default configuration of :class:`BOCDConfig` is used.
    :type config: Optional[BOCDConfig]
    :param callbacks: callbacks, defaults to None
    :type callbacks: Optional[Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]]

    :Note:
     Adapted from the implementation in https://github.com/gwgundersen/bocd.

    :References:

    .. [adams2007bayesian] Adams, Ryan Prescott, and David JC MacKay.
        "Bayesian online changepoint detection."
        arXiv preprint arXiv:0710.3742 (2007).

    :Example:

    >>> from frouros.detectors.concept_drift import BOCD
    >>> import numpy as np
    >>> np.random.seed(seed=31)
    >>> dist_a = np.random.normal(loc=0.2, scale=0.01, size=1000)
    >>> dist_b = np.random.normal(loc=0.8, scale=0.04, size=1000)
    >>> stream = np.concatenate((dist_a, dist_b))
    >>> detector = BOCD()
    >>> for i, value in enumerate(stream):
    ...     _ = detector.update(value=value)
    ...     if detector.drift:
    ...         print(f"Change detected at step {i}")
    ...         break
    Change detected at step 1031
    """  # noqa: E501  # pylint: disable=line-too-long

    config_type = BOCDConfig  # type: ignore

    def __init__(  # noqa: D107
        self,
        config: Optional[BOCDConfig] = None,
        callbacks: Optional[
            Union[BaseCallbackStreaming, list[BaseCallbackStreaming]]
        ] = None,
    ) -> None:
        super().__init__(
            config=config,
            callbacks=callbacks,
        )
        self.additional_vars = {
            "log_r": np.array([[0.0]]),
            "predicted_mean": None,
            "predicted_var": None,
            "log_message": np.array([0.0]),
        }
        self._set_additional_vars_callback()
        self._model = copy.deepcopy(self.config.model)  # type: ignore

    @property
    def log_r(self) -> np.ndarray:
        """Log r getter.

        :return: log r
        :rtype: numpy.ndarray
        """
        return self._additional_vars["log_r"]

    @log_r.setter
    def log_r(self, value: np.ndarray) -> None:
        """Log r setter.

        :param value: value to be set
        :type value: numpy.ndarray
        """
        self._additional_vars["log_r"] = value

    @property
    def predicted_mean(self) -> Optional[float]:
        """Predicted mean getter.

        :return: predicted mean
        :rtype: Optional[float]
        """
        return self._additional_vars["predicted_mean"]

    @predicted_mean.setter
    def predicted_mean(self, value: Optional[float]) -> None:
        """Predicted mean setter.

        :param value: value to be set
        :type value: Optional[float]
        """
        self._additional_vars["predicted_mean"] = value

    @property
    def predicted_var(self) -> Optional[float]:
        """Predicted variance getter.

        :return: predicted variance
        :rtype: Optional[float]
        """
        return self._additional_vars["predicted_var"]

    @predicted_var.setter
    def predicted_var(self, value: Optional[float]) -> None:
        """Predicted variance setter.

        :param value: value to be set
        :type value: Optional[float]
        """
        self._additional_vars["predicted_var"] = value

    @property
    def log_message(self) -> np.ndarray:
        """Log message getter.

        :return: log message
        :rtype: numpy.ndarray
        """
        return self._additional_vars["log_message"]

    @log_message.setter
    def log_message(self, value: np.ndarray) -> None:
        """Log message setter.

        :param value: value to be set
        :type value: numpy.ndarray
        """
        self._additional_vars["log_message"] = value

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1
        current_idx = self.num_instances - 1

        # 3. Evaluate predictive probabilities.
        log_pis = self._model.log_pred_prob(
            idx=self.num_instances,
            value=value,
        )

        # 4. Calculate growth probabilities.
        log_pis_plus_message = log_pis + self.log_message
        log_growth_probs = (
            log_pis_plus_message + self.config.log_1_minus_hazard  # type: ignore
        )

        # 5. Calculate changepoint probability.
        log_cp_prob = logsumexp(
            log_pis_plus_message + self.config.log_hazard  # type: ignore
        )

        # 6. Calculate evidence
        new_log_joint = np.append(log_cp_prob, log_growth_probs)

        # 7. Determine run length distribution.
        # Dynamically expand log_r
        self.log_r = np.concatenate(
            (
                np.pad(
                    array=self.log_r,
                    pad_width=((0, 0), (0, 1)),
                    constant_values=-np.inf,
                ),
                np.expand_dims(new_log_joint - logsumexp(new_log_joint), axis=0),
            ),
            axis=0,
        )

        # 8. Update sufficient statistics.
        self._model.update(
            value=value,
        )

        # Update message.
        self.log_message = new_log_joint

        # 9. Perform prediction.
        probs = np.exp(self.log_r[current_idx, : self.num_instances])
        self.predicted_mean = np.sum(
            probs * self._model.mean_params[: self.num_instances]
        )
        self.predicted_var = np.sum(
            probs * self._model.var_params[: self.num_instances]
        )

        if self.num_instances >= self.config.min_num_instances:
            self.drift = (
                self.log_r[self.num_instances, : self.num_instances + 1].argmax()
                != self.num_instances
            )

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.log_r = np.array([[0.0]])
        self.log_message = np.array([0.0])
        self.predicted_mean = None
        self.predicted_var = None
        self._model = copy.deepcopy(self.config.model)  # type: ignore



"""Base data drift distance based module."""

import abc
from typing import Any, Optional, Tuple, Union

import numpy as np  # type: ignore

from frouros.detectors.data_drift.base import BaseResult
from frouros.detectors.data_drift.streaming.base import (
    BaseDataDriftStreaming,
)


class DistanceResult(BaseResult):
    """Distance result class."""

    def __init__(
        self,
        distance: Union[int, float],
    ) -> None:
        """Init method.

        :param distance: distance value
        :type distance: Union[int, float]
        """
        super().__init__()
        self.distance = distance

    @property
    def distance(self) -> Union[int, float]:
        """Distance value property.

        :return: distance value
        :rtype: Union[int, float]
        """
        return self._distance

    @distance.setter
    def distance(self, value: Union[int, float]) -> None:
        """Distance value setter.

        :param value: value to be set
        :type value: Union[int, float]
        """
        self._distance = value


class BaseDistanceBased(BaseDataDriftStreaming):
    """Abstract class representing a distance based."""

    @abc.abstractmethod
    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        pass

    @abc.abstractmethod
    def _reset(self) -> None:
        pass

    @abc.abstractmethod
    def _update(self, value: Union[int, float]) -> Optional[DistanceResult]:
        pass

    def compare(
        self,
        X: np.ndarray,  # noqa: N803
    ) -> Tuple[Optional[DistanceResult], dict[str, Any]]:
        """Compare detector.

        :param X: data to use to compare the detector
        :type X: np.ndarray
        :return: update result
        :rtype: Tuple[Optional[DistanceResult], dict[str, Any]]
        """
        result = self._compare(X=X)
        return result

    @abc.abstractmethod
    def _compare(
        self,
        X: np.ndarray,  # noqa: N803
    ) -> Tuple[Optional[DistanceResult], dict[str, Any]]:
        pass



"""Callbacks init."""

from .batch import PermutationTestDistanceBased, ResetStatisticalTest
from .streaming import HistoryConceptDrift

__all__ = [
    "HistoryConceptDrift",
    "PermutationTestDistanceBased",
    "ResetStatisticalTest",
]
